unit plotter;

{$mode objfpc}

interface

uses
  Classes, SysUtils, FileUtil, TAGraph, TASeries, TATransformations,
  TAIntervalSources, TATools, LResources, Forms, Controls, Graphics, Dialogs,
  Grids, ShellCtrls, ComCtrls, StdCtrls, ExtCtrls, PairSplitter, EditBtn,
  Buttons, FileCtrl, upascaltz
  , Types, TAChartUtils;

type

  { TPlotterForm }

  TPlotterForm = class(TForm)
    AccumulateCheckBox: TCheckBox;
    BitBtn1: TBitBtn;
    SunSeries: TLineSeries;
    ChartButton: TSpeedButton;
    ChartToolset1ZoomDragTool1: TZoomDragTool;
    Label1: TLabel;
    ResetToLogDirectory: TBitBtn;
    ChartToolset1: TChartToolset;
    ChartToolset1DataPointCrosshairTool1: TDataPointCrosshairTool;
    ChartToolset1PanDragTool1: TPanDragTool;
    ChartToolset1ZoomMouseWheelTool1: TZoomMouseWheelTool;
    EditPositionButton: TBitBtn;
    FixedYCheckBox: TCheckBox;
    readingstext: TLabel;
    MiscGroupBox: TGroupBox;
    readingstextR: TLabel;
    readingstextG: TLabel;
    readingstextB: TLabel;
    BlueSeries: TLineSeries;
    ClearPlotButton: TButton;
    DirectoryEdit3: TDirectoryEdit;
    FilesGroupBox: TGroupBox;
    SettingsButton: TSpeedButton;
    SettingsGroupBox: TGroupBox;
    DefaultGroupBox: TGroupBox;
    HelpButton: TButton;
    Label11: TLabel;
    Label6: TLabel;
    PlotNumberGroup: TRadioGroup;
    PositionEntry: TLabeledEdit;
    ColorSwatch: TShape;
    StringGrid1: TStringGrid;
    ThreeDayCheckBox: TCheckBox;
    TimeRefGroup: TRadioGroup;
    TwilightSeries: TLineSeries;
    DateTimeIntervalChartSource1: TDateTimeIntervalChartSource;
    MPSASSeries2: TLineSeries;
    MPSASSeries3: TLineSeries;
    Chart1: TChart;
    ChartAxisTransformations1: TChartAxisTransformations;
    ChartAxisTransformations1LinearAxisTransform1: TLinearAxisTransform;
    MoonSeries: TLineSeries;
    ClearSeries: TLineSeries;
    GreenSeries: TLineSeries;
    MPSASSeries: TLineSeries;
    RedSeries: TLineSeries;
    TZLocationBox: TComboBox;
    TZRegionBox: TComboBox;
    procedure BitBtn1Click(Sender: TObject);
    procedure ResetToLogDirectoryClick(Sender: TObject);
    procedure ChartToolset1DataPointCrosshairTool1Draw(
      ASender: TDataPointDrawTool);
    procedure EditPositionButtonClick(Sender: TObject);
    procedure ChartButtonClick(Sender: TObject);
    procedure ClearPlotButtonClick(Sender: TObject);
    procedure DirectoryEdit3Change(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure HelpButtonClick(Sender: TObject);
    procedure LogFilesButtonClick(Sender: TObject);
    procedure SettingsButtonClick(Sender: TObject);
    procedure ShellTreeView3Click(Sender: TObject);
    procedure StringGrid1Click(Sender: TObject);
    procedure StringGrid1CompareCells(Sender: TObject; ACol, ARow, BCol,
      BRow: Integer; var Result: integer);
    procedure StringGrid1HeaderClick(Sender: TObject; IsColumn: Boolean;
      Index: Integer);
    procedure TimeRefGroupClick(Sender: TObject);
    //procedure ThreeDayCheckBoxClick(Sender: TObject); {!!! maybe keep this !!!}
    procedure TZLocationBoxChange(Sender: TObject);
    procedure TZRegionBoxChange(Sender: TObject);
  private
    { private declarations }
    procedure ReadINI();
    procedure PlotFile(PlotFile: String);
    procedure AddFilesToList(FilePathName : String);
    procedure FileSelect();
  public
    { public declarations }
  end;


  { Declarations outside the form }
  function FileSizeFormat(bytes:Double):string;
  function AddBackSlash(Instring:String):String;
  function RemoveMultiSlash(Input: String): String;
  function DialogCentered(const aCaption: String;
                    const aMsg: string;
                    DlgType: TMsgDlgType;
                    Buttons: TMsgDlgButtons;
                    HelpCtx: longint;
                    Form: TForm
                    ): TModalResult;

  const
  DefaultSettingsWidth = 475;
  DefaultSettingsheight = 210;

var
  PlotterForm: TPlotterForm;

  MyLatitude: extended = 0.0; //Latitude
  MyLongitude: extended = 0.0; //Longitude
  MyElevation: extended = 0.0; //Elevation

  PlotterINIsection: String;
  AZones: TStringList;
  ptz :TPascalTZ;
  subfix: ansistring; //Used for time zone conversions

  CurrentDirectory: String;
  PlotterModel:Integer;
  PlotType:String=''; //options: Color, empty for standard.

  LastRedDarkness: Real = 0.0;
  LastGreenDarkness: Real = 0.0;
  LastBlueDarkness: Real = 0.0;
  LastClearDarkness: Real = 0.0;

  FullChart:Boolean = False;

implementation
uses
  appsettings
  , dateutils //Required to convert logged UTC string to TDateTime
  , strutils //Required for checking lines in conversion file.
  , moon //required for Moon calculations
  , Math //for float identifier
  , dlheader //For Timezone conversions.
  , LazFileUtils //Necessary for filename extraction
  , worldmap
  , Unit1
  ;

{ TPlotterForm }

procedure TPlotterForm.FormCreate(Sender: TObject);
begin
   { Initialize required variables. }
   AZones:=TStringList.Create;
   ptz := TPascalTZ.Create();

  ReadINI();

  //Chart timestamps
  DateTimeIntervalChartSource1.DateTimeFormat:='hh:mm'+#13#10+'DD-mmm'+#13#10+'yyyy';

end;

{ Usage instructions }
procedure TPlotterForm.HelpButtonClick(Sender: TObject);
begin
  DialogCentered(
    'Navigation help',
    'Zoom : Draw box from top left corner to lower right corner with LEFT mouse button.' + sLineBreak +
    'Reset: Click LEFT mouse button to reset to full view.'+ sLineBreak +
    'Pan: Click and drag RIGHT mouse button to pan chart.',
    mtInformation,
    [mbOK],
    0,
    PlotterForm
    );
end;

procedure TPlotterForm.LogFilesButtonClick(Sender: TObject);
begin
end;

procedure TPlotterForm.SettingsButtonClick(Sender: TObject);
begin
  if (FullChart) then begin
    FilesGroupBox.Width:=(PlotterForm.Width - SettingsButton.Width-DefaultSettingsWidth-10);
    SettingsGroupBox.Height:=DefaultSettingsheight;
    ChartButton.Down:=False;
    SettingsButton.down:=True;
    FullChart:=False;
  end;
end;

procedure TPlotterForm.ChartButtonClick(Sender: TObject);
begin
  if (not FullChart) then begin
    FilesGroupBox.Width:=0;
    SettingsGroupBox.Height:=0;
    ChartButton.Down:=True;
    SettingsButton.down:=False;
    FullChart:=True;
  end;
end;

procedure TPlotterForm.EditPositionButtonClick(Sender: TObject);
begin
  worldmap.FormWorldmap.CalledBy:='Plotter';
  worldmap.FormWorldmap.ShowModal;
end;

{Clear plot }
procedure TPlotterForm.ClearPlotButtonClick(Sender: TObject);
begin
 MPSASSeries.Clear;
 MPSASSeries2.Clear;
 MPSASSeries3.Clear;
 RedSeries.Clear;
 GreenSeries.Clear;
 BlueSeries.Clear;
 ClearSeries.Clear;
 SunSeries.Clear;
 MoonSeries.Clear;
 TwilightSeries.Clear;
end;

{Reset plot file directory to logs file directory}
procedure TPlotterForm.ResetToLogDirectoryClick(Sender: TObject);
begin
  DirectoryEdit3.Text:=appsettings.LogsDirectory+'/';
  vConfigurations.WriteString(PlotterINIsection,'TreeViewPath',DirectoryEdit3.Text);
end;

procedure TPlotterForm.BitBtn1Click(Sender: TObject);
begin
  Chart1.ZoomFull();
end;

{ Display the cursor position value }
procedure TPlotterForm.ChartToolset1DataPointCrosshairTool1Draw(
  ASender: TDataPointDrawTool);
var
    RedValue, GreenValue, BlueValue:Byte;
    Red, Green, Blue, Maximum, Minimum, Range: Float;
begin
  if (PlotType='Color') then begin //color model
    Red:=RedSeries.GetYValue(ASender.PointIndex);
    Green:=GreenSeries.GetYValue(ASender.PointIndex);
    Blue:=BlueSeries.GetYValue(ASender.PointIndex);
    readingstextR.Caption:= FloatToStr(Red);
    readingstextG.Caption:= FloatToStr(Green);
    readingstextB.Caption:= FloatToStr(Blue);
    {Make swatch ...}
    { - Invert darkness value to get brightness values }
    Red:= -1 * Red;
    Green:= -1 * Green;
    Blue:= -1 * Blue;
    { - Find Maximum }
    Maximum:=abs(Max(Max(Red,Green),Blue));
    { - Find Minimum }
    Minimum:=abs(Min(Min(Red,Green),Blue));
    { - Apply offset}
    Red:=Red+Minimum;
    Green:=Green+Minimum;
    Blue:=Blue+Minimum;
    { - Normalize values }
    Range:=abs(Maximum - Minimum);
    RedValue:= round(Red*255/Range);
    GreenValue:=round(Green*255/Range);
    BlueValue:=round(Blue*255/Range);
    { - Colorize the swatch }
    ColorSwatch.Brush.Color:=RGBToColor(RedValue, GreenValue, BlueValue);
  end
  else
    readingstext.Caption:= Format('%.2f mpsas',[MPSASSeries.GetYValue(ASender.PointIndex)]);
end;

{User has selected a directory of files to choose plotfile from}
procedure TPlotterForm.DirectoryEdit3Change(Sender: TObject);
begin
  AddFilesToList(DirectoryEdit3.Directory);
end;

{ Populate form from INI file }
procedure TPlotterForm.ReadINI();
var
  pieces: TStringList;

begin

   pieces := TStringList.Create;
   pieces.Delimiter := ',';
   pieces.StrictDelimiter := False; //Parse spaces also

   PlotterINIsection:='Plotter:';

   { Read previously selected file path }
   DirectoryEdit3.Text:=vConfigurations.ReadString(PlotterINIsection, 'TreeViewPath', appsettings.LogsDirectory+'/');

   { Pull Timezone information from INI file if it exists.}
   TZRegionBox.Text:= vConfigurations.ReadString(PlotterINIsection,'Local region');
   if (FileExists(appsettings.TZDirectory+TZRegionBox.Text) and (length(TZRegionBox.Text)>0))then begin
       ptz.Destroy;
       ptz := TPascalTZ.Create();
       ptz.ParseDatabaseFromFile(appsettings.TZDirectory+TZRegionBox.Text);
       ptz.GetTimeZoneNames(AZones,true);  //only geo name = true, does not show short names
       TZLocationBox.Items.AddStrings(AZones);
     end;

   { Read the previously recorded entries. }
   TZLocationBox.Text:= vConfigurations.ReadString(PlotterINIsection,'Local time zone');

   PositionEntry.Text:=vConfigurations.ReadString(PlotterINIsection,'Position');
   //Parse location
   pieces.DelimitedText := PositionEntry.Text;
   if pieces.Count>1 then begin
     MyLatitude:=StrToFloatDef(pieces.Strings[0],0);
     MyLongitude:=StrToFloatDef(pieces.Strings[1],0);
   end else begin
     MyLatitude:=0;
     MyLongitude:=0;
   end;
   //Parse elevation
   if pieces.Count>2 then begin
     MyElevation:=StrToFloatDef(pieces.Strings[2],0);
   end else begin
     MyElevation:=0;
   end;

   if Assigned(pieces) then FreeAndNil(pieces);

end;
procedure TPlotterForm.PlotFile(PlotFile: String);
var
    File1: TextFile;
    Str: String;
    pieces: TStringList;
    i: Integer; //general purpose counter
    MSASField: Integer = -1; //Field that contains the MSAS variable, -1 = not defined  yet.
    ScaleField: Integer = -1; //Field that contains the Scale variable, -1 = not defined  yet.
    ColorField: Integer = -1; //Field that contains the Color variable, -1 = not defined  yet.

    MoonElevation: extended = 0.0;
    MoonAzimuth: extended = 0.0;

    SunElevation: extended = 0.0;
    SunAzimuth: extended = 0.0;

    LocalRecordTime,UTCRecordTime,PlotRecordTime,TwilightStamp :TDateTime;
    LastUTCDate:TDateTime;
    Darkness: Float;
    ThisLocation: String;
    FirmwareString: String;
    DateString:String;
    FS: TFormatSettings;

procedure PositionParse();
begin
   //Get Lat, Lon for Moon calculations
   pieces.Delimiter := ',';
   //Separate the fields of the lat,lon.
   pieces.DelimitedText := PositionEntry.Text;

   if (pieces.Count > 1) then begin
     MyLatitude:=StrToFloat(pieces.Strings[0]);
     if ((MyLatitude<-90.0) or (MyLatitude>90.0)) then begin
         MessageDlg('Latitude Error',
           'Error: Latitude ' + FloatToStr(MyLatitude)+' out of range.',
           mtError,
           [mbOK],0);
     end;
     MyLongitude:=StrToFloat(pieces.Strings[1]);
     if ((MyLongitude<-180.0) or (MyLongitude>180.0)) then begin
         MessageDlg('Longitude Error',
           'Error: Longitude ' + FloatToStr(MyLongitude)+' out of range.',
           mtError,
           [mbOK],0);
     end;
   end;
end;

begin
   pieces := TStringList.Create;
   AssignFile(File1, PlotFile);

   FS := DefaultFormatSettings;
   FS.DateSeparator := '/';
   FS.ShortDateFormat := 'yyyy/mm/dd';
   FS.ShortTimeFormat := 'hh:mm:ss';

   { Save the TZ region selection. }
   vConfigurations.WriteString(SerialINIsection,'Local region',TZRegionBox.Text);

   { Read the region database table. }
   ptz.Destroy;
   ptz := TPascalTZ.Create();
   ptz.ParseDatabaseFromFile(appsettings.TZDirectory+TZRegionBox.Text);
   Azones.Clear;
   ptz.GetTimeZoneNames(AZones,true); //only geo name = true

   case ExtractFileExt(PlotFile) of

   //SQM Reader files
   '.txt':begin
     {$I+}
     try

       //Default to no model selected
       PlotterModel:=0;

       //Default to no plot type selected
       PlotType:='';

       Reset(File1);
       repeat
         // Read one line at a time from the file.
         Readln(File1, Str);

         //Check for comment lines
         //Skip improper dates (like 1985)
         if (AnsiStartsStr('Produced by SQM Reader',Str) or AnsiStartsStr('Year',Str)) then begin

           //Ignore these lines

         end else begin
             //Separate the fields of the record.
            pieces.Delimiter := ',';
            pieces.DelimitedText := Str;

           //Separate the fields of the record.
           pieces.DelimitedText := Str;


           //Set MPSAS field
           MSASField:=2;

              DateString:=pieces.Strings[0]+ ' ' +pieces.Strings[1];
              //LocalRecordTime:=ScanDateTime('yyyy/mm/dd hh:nn:ss',DateString);
              LocalRecordTime:= StrToDateTime(DateString, FS);

              PlotRecordTime:=LocalRecordTime;

             //Update Chart
             Darkness:= StrToFloatDef(AnsiLeftStr(pieces.Strings[MSASField],Length(pieces.Strings[MSASField])),0);

               if PlotNumberGroup.ItemIndex=0 then
                  MPSASSeries.AddXY(PlotRecordTime, Darkness);
               if PlotNumberGroup.ItemIndex=1 then
                  MPSASSeries2.AddXY(PlotRecordTime, Darkness);
               if PlotNumberGroup.ItemIndex=2 then
                  MPSASSeries3.AddXY(PlotRecordTime, Darkness);

         end;//end of checking for data

       until(EOF(File1)); // EOF(End Of File) The the program will keep reading new lines until there is none.
       CloseFile(File1);

     except
       on E: EInOutError do
         MessageDlg('Error', 'File handling error occurred. Details: '+E.ClassName+'/'+E.Message, mtError, [mbOK],0);
       on E: Exception do
         ShowMessage( 'Error: '+ E.ClassName + #13#10 + '>' + str + '<' + #13#10 + E.Message );

     end;



   end;

   //Old log files from Perl script
   '.log': begin
            {$I+}
            try

              //Get Lat, Lon for Moon calculations
              pieces.Delimiter := ',';
              //Separate the fields of the lat,lon.
              pieces.DelimitedText := PositionEntry.Text;

              MyLatitude:=StrToFloat(pieces.Strings[0]);
              if ((MyLatitude<-90.0) or (MyLatitude>90.0)) then begin
                  MessageDlg('Latitude Error',
                    'Error: Latitude ' + FloatToStr(MyLatitude)+' out of range.',
                    mtError,
                    [mbOK],0);
              end;
              MyLongitude:=StrToFloat(pieces.Strings[1]);
              if ((MyLongitude<-180.0) or (MyLongitude>180.0)) then begin
                  MessageDlg('Longitude Error',
                    'Error: Longitude ' + FloatToStr(MyLongitude)+' out of range.',
                    mtError,
                    [mbOK],0);
              end;

               //Prevent errors of "time does not exist in"
               ptz.DetectInvalidLocalTimes:=False;

              Reset(File1);
              pieces.Delimiter := ' ';
              repeat
                // Read one line at a time from the file.
                Readln(File1, Str);

                //Separate the fields of the record.
                pieces.DelimitedText := Str;

                //Make sure there are enough fields
                if (pieces.Count<>3) then begin
                    MessageDlg('Error', 'Incorrect number of fields in record. Expected 3, got: '+ IntToStr(pieces.Count), mtError, [mbOK],0);
                    break;
                  end
                else begin
                  //parse the fields, and convert as necessary.
                  //Convert UTC string 'YYYY-MM-DDTHH:mm:ss.fff' into TDateTime
                  LocalRecordTime:=ScanDateTime('yyyymmddhh:nn:ss',LazFileUtils.ExtractFileNameOnly(PlotFile) + pieces.Strings[0]);

                  UTCRecordTime := ptz.LocalTimeToGMT(LocalRecordTime,TZLocationBox.Text);

                  //Convert local time to UTC for Moon calculation

                  //Update Chart
                  Darkness:= StrToFloatDef(AnsiLeftStr(pieces.Strings[1],Length(pieces.Strings[1])-1),0); //remove trailing m from mpsas
                  MPSASSeries.AddXY(LocalRecordTime, Darkness);
                  //Calculate Moon position
                  //Change sign for Moon calculations
                  Moon_Position_Horizontal(
                                          StrToDateTime(DateTimeToStr(UTCRecordTime)),
                                          -1.0*MyLongitude,
                                          MyLatitude,
                                          MoonElevation,
                                          MoonAzimuth);
                  MoonSeries.AddXY(LocalRecordTime,MoonElevation);
                  //Calculate Sun position
                  //Change sign for Sun calculations
                  Sun_Position_Horizontal(StrToDateTime(DateTimeToStr(UTCRecordTime)),
                                          -1.0*MyLongitude,
                                          MyLatitude,
                                          SunElevation,
                                          SunAzimuth);
                  SunSeries.AddXY(LocalRecordTime,SunElevation);

                end;//End of checking number of fields in record.

              until(EOF(File1)); // EOF(End Of File) The the program will keep reading new lines until there is none.
              CloseFile(File1);



            except
              on E: EInOutError do begin
               MessageDlg('Error', 'File handling error occurred. Details: '+E.ClassName+'/'+E.Message, mtError, [mbOK],0);
              end;

            end;
   end;

   //New dat files from
   '.dat': begin
             {$I+}
             try

               PositionParse();

               //Get the location for use in this procedure
               ThisLocation := TZLocationBox.Text;


               //Prevent errors of "time does not exist in"
               ptz.DetectInvalidLocalTimes:=False;

               //Default to no model selected
               PlotterModel:=0;

               //Default to no plot type selected
               PlotType:='';

               Reset(File1);
               repeat
                 // Read one line at a time from the file.
                 Readln(File1, Str);

                 //Check for comment lines
                 //Skip improper dates (like 1985)
                 if (AnsiStartsStr('#',Str) or AnsiStartsStr('1985',Str)) then begin

                   //Parse position definition line
                   if (AnsiStartsStr('# Position',str)) then begin
                      pieces.Delimiter := ':';
                      pieces.StrictDelimiter := True; //Do not parse spaces also
                      pieces.DelimitedText := Str;
                      if (pieces.Count>1) then begin
                        if Trim(pieces.Strings[1]) <>'' then begin
                          PositionEntry.Text:=Trim(pieces.Strings[1]);
                          PositionParse();
                        end;
                      end;
                   end;

                   //These fields do not really matter since the local and UTC times are in the records already
                   //Parse Local timezone region line
                   //if (AnsiStartsStr('# Position',str)) then begin
                   //   pieces.Delimiter := ':';
                   //   pieces.StrictDelimiter := True; //Do not parse spaces also
                   //   pieces.DelimitedText := Str;
                   //   if (pieces.Count>1) then begin
                   //     if Trim(pieces.Strings[1]) <>'' then begin
                   //       PositionEntry.Text:=Trim(pieces.Strings[1]);
                   //       PositionParse();
                   //     end;
                   //   end;
                   //
                   //   { Pull Timezone information from INI file if it exists.}
                   //   TZRegionBox.Text:= vConfigurations.ReadString(SerialINIsection,'Local region');
                   //   if (FileExists(appsettings.TZDirectory+TZRegionBox.Text) and (length(TZRegionBox.Text)>0))then
                   //     begin
                   //       ptz.Destroy;
                   //       ptz := TPascalTZ.Create();
                   //       ptz.ParseDatabaseFromFile(appsettings.TZDirectory+TZRegionBox.Text);
                   //       ptz.GetTimeZoneNames(AZones,true);  //only geo name = true, does not show short names
                   //       TZLocationBox.Items.AddStrings(AZones);
                   //     end;
                   //end;

                   //Parse Local timezone name line
                   // if (AnsiStartsStr('# Position',str)) then begin
                   //   pieces.Delimiter := ':';
                   //   pieces.StrictDelimiter := True; //Do not parse spaces also
                   //   pieces.DelimitedText := Str;
                   //   if (pieces.Count>1) then begin
                   //     if Trim(pieces.Strings[1]) <>'' then begin
                   //       PositionEntry.Text:=Trim(pieces.Strings[1]);
                   //       PositionParse();
                   //     end;
                   //   end;
                   //
                   //   { Read the previously recorded entries. }
                   //   TZLocationBox.Text:= vConfigurations.ReadString(SerialINIsection,'Local time zone');
                   //end;

                   //Parse the model type
                   if (AnsiStartsStr('# SQM firmware version:',str)) then begin
                      pieces.Delimiter := ':';
                      pieces.StrictDelimiter := True; //Do not parse spaces also
                      pieces.DelimitedText := Str;
                      if (pieces.Count>1) then begin
                        FirmwareString:=pieces.Strings[1];
                        pieces.Delimiter := '-';
                        pieces.StrictDelimiter := True; //Do not parse spaces also
                        pieces.DelimitedText := FirmwareString;
                         if (pieces.Count=3) then begin
                           PlotterModel:= StrToIntDef(pieces.Strings[1],0);
                         end;
                      end;
                   end;

                   //Parse field definition line
                   if (AnsiContainsStr(str,'UTC') and AnsiContainsStr(str,'Local') and AnsiContainsStr(str,'MSAS')) then begin
                     pieces.Delimiter := ',';
                     pieces.StrictDelimiter := True; //Do not parse spaces also
                     pieces.DelimitedText := Str;

                     { Check which fiels the reading (MSAS) is in }
                     for i:=0 to pieces.Count-1 do begin
                       if AnsiContainsStr(pieces.Strings[i],'MSAS') then MSASField:=i;
                       if AnsiContainsStr(pieces.Strings[i],'Scale') then ScaleField:=i;
                       if AnsiContainsStr(pieces.Strings[i],'Color') then ColorField:=i;
                     end;
                     if ((PlotterModel=4) and (ScaleField>0) and (ColorField>0)) then PlotType:='Color';
                   end;

                 end else begin
                     //Separate the fields of the record.
                    pieces.Delimiter := ';';
                    pieces.DelimitedText := Str;

                   //Separate the fields of the record.
                   pieces.DelimitedText := Str;


                   //Make sure there are enough fields
                   if (MSASField=-1) then begin

                       DialogCentered(
                         'Error',
                         'No MSAS field defined in file',
                         mtError,
                         [mbOK],
                         0,
                         PlotterForm
                        );

                       break;
                     end
                     //if (pieces.Count<>5) then begin
                     //    MessageDlg('Error', 'Incorrect number of fields in record. Expected 5, got: '+ IntToStr(pieces.Count), mtError, [mbOK],0);
                     //    break;
                     //  end
                   else begin

                     UTCRecordTime  :=ScanDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz',pieces.Strings[0]);

                     //Time reference for plotting.
                     if TimeRefGroup.ItemIndex=0 then
                        PlotRecordTime:=UTCRecordTime
                     else begin
                        LocalRecordTime:=ScanDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz',pieces.Strings[1]);
                        PlotRecordTime:=LocalRecordTime;
                     end;

                     //Calculate civil sunrise
                     //only check if last date differs
                     if CompareDate(UTCRecordTime,LastUTCDate)<>0 then begin
                       //write(FormatDateTime('tt',ptz.GMTToLocalTime(Morning_Twilight_Civil(UTCRecordTime,
                       //   MyLatitude, -1.0 * MyLongitude), ThisLocation, subfix))+' ');
                       //   MyLatitude, -1.0 * MyLongitude), ThisLocation, subfix)));
                       {Civil Morning stamp}
                       TwilightStamp:=ptz.GMTToLocalTime(Morning_Twilight_Civil(UTCRecordTime,
                          MyLatitude, -1.0 * MyLongitude), ThisLocation, subfix);
                       Sun_Position_Horizontal(StrToDateTime(DateTimeToStr(UTCRecordTime)),
                                               -1.0*MyLongitude,
                                               MyLatitude,
                                               SunElevation,
                                               SunAzimuth);
                       TwilightSeries.AddXY(TwilightStamp,SunElevation,'M',clYellow);

                       {Civil Evening stamp}
                       TwilightStamp:=ptz.GMTToLocalTime(Evening_Twilight_Civil(UTCRecordTime,
                          MyLatitude, -1.0 * MyLongitude), ThisLocation, subfix);
                       Sun_Position_Horizontal(StrToDateTime(DateTimeToStr(UTCRecordTime)),
                                                  -1.0*MyLongitude,
                                                  MyLatitude,
                                                  SunElevation,
                                                  SunAzimuth);
                       TwilightSeries.AddXY(TwilightStamp,SunElevation,'E',clBlue);


                     end;
                     LastUTCDate:=UTCRecordTime;

                     //Update Chart
                     Darkness:= StrToFloatDef(AnsiLeftStr(pieces.Strings[MSASField],Length(pieces.Strings[MSASField])),0);

                     if (PlotType='Color') then begin {Color model}
                       writeln(format('scalefield=%d colorfield=%d',[ScaleField,ColorField]));
                       case StrToIntDef(pieces.Strings[ColorField],0) of
                         0: LastRedDarkness:=Darkness;
                         1: LastBlueDarkness:=Darkness;
                         2: LastClearDarkness:=Darkness;
                         3: LastgreenDarkness:=Darkness;
                       end;
                       RedSeries.AddXY(PlotRecordTime, LastRedDarkness);
                       GreenSeries.AddXY(PlotRecordTime, LastgreenDarkness);
                       BlueSeries.AddXY(PlotRecordTime, LastBlueDarkness);
                       ClearSeries.AddXY(PlotRecordTime, LastClearDarkness);
                     end
                     else begin {non color models}
                       if PlotNumberGroup.ItemIndex=0 then
                          MPSASSeries.AddXY(PlotRecordTime, Darkness);
                       if PlotNumberGroup.ItemIndex=1 then
                          MPSASSeries2.AddXY(PlotRecordTime, Darkness);
                       if PlotNumberGroup.ItemIndex=2 then
                          MPSASSeries3.AddXY(PlotRecordTime, Darkness);

                     end;

                     //Calculate Moon position
                     //Change sign for Moon calculations
                     Moon_Position_Horizontal(
                                             StrToDateTime(DateTimeToStr(UTCRecordTime)),
                                             -1.0*MyLongitude,
                                             MyLatitude,
                                             MoonElevation,
                                             MoonAzimuth);
                     MoonSeries.AddXY(PlotRecordTime,MoonElevation);

                     //Calculate Sun position
                     //Change sign for Sun calculations
                     Sun_Position_Horizontal(StrToDateTime(DateTimeToStr(UTCRecordTime)),
                                             -1.0*MyLongitude,
                                             MyLatitude,
                                             SunElevation,
                                             SunAzimuth);
                     SunSeries.AddXY(PlotRecordTime,SunElevation);

                   end;//End of checking number of fields in record.

                 end;//end of checking for data

               until(EOF(File1)); // EOF(End Of File) The the program will keep reading new lines until there is none.
               CloseFile(File1);

             except
               on E: EInOutError do
                 MessageDlg('Error', 'File handling error occurred. Details: '+E.ClassName+'/'+E.Message, mtError, [mbOK],0);
               on E: Exception do
                 ShowMessage( 'Error: '+ E.ClassName + #13#10 + '>' + str + '<' + #13#10 + E.Message );

             end;

   end;

   else begin end;

   end;

end;

{Save selected directory}
procedure TPlotterForm.ShellTreeView3Click(Sender: TObject);
begin
  vConfigurations.WriteString(PlotterINIsection,'TreeViewPath',DirectoryEdit3.Text);
end;

{ User requests file to be plotted }
procedure TPlotterForm.StringGrid1Click(Sender: TObject);
begin
  FileSelect();
end;

{ Custom sorting routine for string/number/date}
procedure TPlotterForm.StringGrid1CompareCells(Sender: TObject; ACol, ARow,
  BCol, BRow: Integer; var Result: integer);
begin
  case StringGrid1.SortColumn of
  0: Result:=CompareText(StringGrid1.Cells[ACol,ARow],StringGrid1.Cells[BCol,BRow]);
  1: begin //numeric sort for column 1 (numerical)
      // Result will be either <0, =0, or >0 for normal order.
        result := StrToIntDef(StringGrid1.Cells[ACol,ARow],0)-StrToIntDef(StringGrid1.Cells[BCol,BRow],0);
     end;
  2: begin //Date compare
      try  //In case date is unreadable or empty
        result := DateTimeToUnix(StrToDateTime(StringGrid1.Cells[ACol,ARow],FDateSettings))
               - DateTimeToUnix(StrToDateTime(StringGrid1.Cells[BCol,BRow],FDateSettings));
      except
      end;
     end;
  end;

  // For inverse order, just negate the result (eg. based on grid's SortOrder).
  if StringGrid1.SortOrder = soDescending then
     result := -result;

end;

{ User requests sorting }
procedure TPlotterForm.StringGrid1HeaderClick(Sender: TObject;
  IsColumn: Boolean; Index: Integer);
begin
   if IsColumn then
    StringGrid1.SortColRow(True, Index);
end;

procedure TPlotterForm.TimeRefGroupClick(Sender: TObject);
begin
  {Update plot when Time reference has been changed}
  FileSelect();
end;

{!!! Maybe keep this !!!}
//procedure TPlotterForm.ThreeDayCheckBoxClick(Sender: TObject);
//begin
// ShellListView1Click(nil);
//end;

procedure TPlotterForm.TZLocationBoxChange(Sender: TObject);
begin
 { Save the TZ selection. }
 vConfigurations.WriteString(PlotterINIsection,'Local time zone',TZLocationBox.Text);
end;

procedure TPlotterForm.TZRegionBoxChange(Sender: TObject);
begin
 { Clear out location because region has changed. }
 TZLocationBox.Text:='';

 { Save the TZ region selection. }
 vConfigurations.WriteString(PlotterINIsection,'Local region',TZRegionBox.Text);

  { Fill up timezone names }
  if (FileExists(appsettings.TZDirectory+TZRegionBox.Text) and (length(TZRegionBox.Text)>0))then begin
     ptz.Destroy;
     ptz := TPascalTZ.Create();
     ptz.ParseDatabaseFromFile(appsettings.TZDirectory+TZRegionBox.Text);
     ptz.GetTimeZoneNames(AZones,true);  //only geo name = true, does not show short names
     TZLocationBox.Items.AddStrings(AZones);
  end;

 { Read the previously recorded entries. }
 TZLocationBox.Text:= vConfigurations.ReadString(PlotterINIsection,'Local time zone');

end;

procedure TPlotterForm.AddFilesToList(FilePathName : String);
var
  D, Row, FileSize : integer;
  FileName, FilePath, FileDate : string;
  SearchRec2 : TSearchRec;
  FileListDat : TStringList;
begin
  FileListDat := TStringList.Create;
  FileName := ExtractFileName(FilePathName);
  FilePath := ExtractFilePath(FilePathName);
  Row := 1; {Starting row}
  FileSize := 0;

  FileListDat := FindAllFiles(FilePath, '*.*', true, faDirectory); //was .dat only

  Row := 1; {Starting row}
  for D := 0 to FileListDat.Count -1 do begin
    if FindFirstUtf8(FilePath, faAnyFile and faDirectory,SearchRec2)=0 then begin
      repeat
        With SearchRec2 do begin
          FileName := ExtractFileName(FileListDat.Strings[D]);
          FileSize := FileSizeUtf8(FileListDat.Strings[D]);
          FileDate:=DateTimeToStr(FileDateToDateTime(FileAge(FileListDat.Strings[D])),FDateSettings);
          if (FileName<>'') then begin
             StringGrid1.InsertRowWithValues(Row,[FileName, IntToStr(FileSize), FileDate ]);
             Row := Row + 1;
          end;
         end;
       until FindNextUtf8(SearchRec2) <> 0;
     end;
     FindCloseUtf8(SearchRec2);
   end;
   StringGrid1.SortColRow(true, 1,1,StringGrid1.RowCount-1);
   FileListDat.Free;
end;

procedure TPlotterForm.FileSelect();
var
    InFilename,SelectedDateFile, PreviousDateFile,NextDateFile:String;
    File1: TextFile;
    Str: String;
    pieces: TStringList;

    MoonElevation: extended = 0.0;
    MoonAzimuth: extended = 0.0;

    UTCRecord, SelectedDate, PreviousDate, NextDate :TDateTime;
    ComposeString: String;
begin

  CurrentDirectory:=AddBackSlash(DirectoryEdit3.Directory);

  pieces := TStringList.Create;

  { New plot for each file clicked }
  if not AccumulateCheckBox.Checked then begin
    Chart1.ZoomFull(True);
    ClearPlotButtonClick(nil);
  end;

  InFilename:= CurrentDirectory+StringGrid1.Cells[0,StringGrid1.Row];

  if ((not FileExists(InFilename)) or (InFilename='')) then exit;

  PlotterForm.Caption:='Potter : '+InFilename;

  {.txt files (from SQM Reader) }
  if (ExtractFileExt(InFilename)= '.txt') then begin

     PlotFile(InFilename);

  end
  {other ext like .dat or .log should have date in filename}
  else begin

    { Date should be encoded into filename, otherwise do not plot }
    try
      SelectedDate:=ScanDateTime('yyyymmdd',LazFileUtils.ExtractFileNameOnly(InFilename));
      SelectedDateFile:=InFilename;

      //Check day before and after for more plotting
      if ThreeDayCheckBox.Checked then begin
        PreviousDate:=IncDay(SelectedDate,-1);
        PreviousDateFile:=CurrentDirectory+FormatDateTime('yyyymmdd',PreviousDate)+ExtractFileExt(InFilename);

        NextDate:=IncDay(SelectedDate);
        NextDateFile:=CurrentDirectory+FormatDateTime('yyyymmdd',NextDate)+ExtractFileExt(InFilename);

        PlotFile(PreviousDateFile);
        PlotFile(SelectedDateFile);
        PlotFile(NextDateFile);
      end else begin
        PlotFile(SelectedDateFile);
      end;

    except
      on E: Exception do
        ShowMessage( 'Error: '+ E.ClassName  + #13#10 + E.Message + #13#10 + 'Could not get date from filename.');
    end;


  end;
//.............................................................................
//.............................................................................

 exit;

  {$I+}
  try
    Reset(File1);

    repeat
      // Read one line at a time from the file.
      Readln(File1, Str);
      // Get location data from header.
      if AnsiStartsStr('# Position',Str) then begin
          //Prepare for parsing.
          pieces.Delimiter := ',';
          pieces.StrictDelimiter := True; //Do not parse spaces.
          //Remove comment from beginning of line.
          pieces.DelimitedText := AnsiRightStr(Str,length(Str) - RPos(':',Str));
          if (pieces.Count<2) then begin
              MessageDlg('Error', 'No location data found in header.', mtError, [mbOK],0);
              MessageDlg('Location Error',
                'Error: No location data found in header.',
                mtError,
                [mbOK],0);
              //LatitudeDisplay.Text:='nul';
              //LongitudeDisplay.Text:='nul';
              //break;
            end
          else
            begin
             //MyLatitude:=StrToFloat(pieces.Strings[0]);
             //if ((MyLatitude<-90.0) or (MyLatitude>90.0)) then begin
             //    writeln('Error: Latitude ',MyLatitude,' out of range');
             //    break;
             //  end;
             //MyLongitude:=StrToFloat(pieces.Strings[1]);
             //if ((MyLongitude<-180.0) or (MyLongitude>180.0)) then begin
             //    writeln('Error: Longitude ',MyLongitude,' out of range');
             //    break;
             //  end;
             // LatitudeDisplay.Text:=Format('%.3f',[MyLatitude]);
             // LongitudeDisplay.Text:=Format('%.3f',[MyLongitude]);
            end;
        end;

      //Get and modify record header string
      //if AnsiStartsStr('# YYYY-MM-DDTHH:mm:ss.fff;',Str) then
      //  Writeln('UTC '+AnsiRightStr(Str,length(Str) - 2)+';MoonPhaseDeg;MoonElevDeg;MoonIllum%');

      //Ignore comment lines which have # as first character.
      if not AnsiStartsStr('#',Str) then begin
          //Separate the fields of the record.
          pieces.Delimiter := ';';
          pieces.DelimitedText := Str;

          //Make sure there are enough fields
          if ((pieces.Count<5) or (pieces.Count>6)) then begin
              MessageDlg('Error', 'Incorrect number of fields in record.', mtError, [mbOK],0);
              break;
            end
          else
          begin

          //parse the fields, and convert as necessary.
          //Convert UTC string 'YYYY-MM-DDTHH:mm:ss.fff' into TDateTime
          UTCRecord:=ScanDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz',pieces.Strings[0]);
          //writeln(FormatDateTime('yyyy-mm-dd"T"hh:nn:ss.zzz',UTCRecord));

          //Calculate Moon position
          //Change sign for Moon calculations
          Moon_Position_Horizontal(
                                   StrToDateTime(DateTimeToStr(UTCRecord)),
                                   -1.0*MyLongitude,
                                   MyLatitude,
                                   MoonElevation,
                                   MoonAzimuth);

          //Prepare string for output.
          ComposeString:= Str + ';'
                         //Moon Phase angle (0 to 180 degrees).
                         + Format('%.1f;',[moon_phase_angle(StrToDateTime(DateTimeToStr(UTCRecord)))])

                         //Moon elevation (positive = above horizon, negative = below horizon).
                         + Format('%.3f;',[MoonElevation])

                         //Moon illumination pecent.
                         + Format('%.1f;',[current_phase(StrToDateTime(DateTimeToStr(UTCRecord)))*100.0])
          ;

          //WriteLn(ComposeString);
          end;//End of checking number of fields in record.

        end;
    until(EOF(File1)); // EOF(End Of File) The the program will keep reading new lines until there is none.
    CloseFile(File1);
  except
    on E: EInOutError do begin
     MessageDlg('Error', 'File handling error occurred. Details: '+E.ClassName+'/'+E.Message, mtError, [mbOK],0);
    end;
  end;

end;

function FileSizeFormat(bytes:Double):string;
begin

  // below 1024 bytes
  if bytes < 1024 then
    Result:=floattostr(bytes) + 'b'

  // kilobytes; 1kb = 1024bytes
  else if (bytes >= 1024) and (bytes < 1024**2) then
    Result:=FormatFloat('0.00', bytes/1024)+'kb'

  // megabytes; 1mb = 1024*1024bytes
  else if (bytes >= 1024**2) and (bytes < 1024**3) then
    Result:=FormatFloat('0.00', bytes/1024/1024)+'Mb'

  // gigabytes; 1gb = 1024*1024*1024bytes
  else if (bytes >= 1024**3) and (bytes < 1024**4) then
    Result:=FormatFloat('0.00', bytes/1024/1024/1024)+'Gb'

  // for everything above, we show it in terrabytes
  else
    Result:=FormatFloat('0.00', bytes/1024/1024/1024/1024)+'Tb';

end;

{ TODO : change slash for Windows }
function AddBackSlash(Instring:String):String;
begin
  Result:=Instring+'/';
end;

function RemoveMultiSlash(Input: String): String;
begin
  Result := Copy(Input,1,2) + StringReplace(Copy(Input,3,Length(Input)),'\\','\',[rfReplaceAll]);
  Result := Copy(Input,1,2) + StringReplace(Copy(Input,3,Length(Input)),'///','/',[rfReplaceAll]);
  Result := Copy(Input,1,2) + StringReplace(Copy(Input,3,Length(Input)),'//','/',[rfReplaceAll]);
end;


function DialogCentered(const aCaption: String;
                  const aMsg: string;
                  DlgType: TMsgDlgType;
                  Buttons: TMsgDlgButtons;
                  HelpCtx: longint;
                  Form: TForm
                  ): TModalResult;
var f: TForm;
begin
  Result := 0;
  f:=CreateMessageDialog(aMsg, DlgType, Buttons);
  try
    f.Caption:=aCaption;
    f.Position:=poDesigned;
    f.Left:=PlotterForm.Left+(PlotterForm.Width div 2)-f.Width div 2;
    f.Top:=PlotterForm.Top + (PlotterForm.Height div 2)-f.Height div 2;

    Result:=f.ShowModal;
  finally
    f.Free;
  end;
end;

initialization
  {$I plotter.lrs}

end.

